@@ -297,7 +297,7 @@ class XQuerySet
 
 		qp->set_database(*db);
 	
-		Xapian::Query * q = new Xapian::Query(qp->parse_query(s));
+		Xapian::Query * q = new Xapian::Query(qp->parse_query(s,Xapian::QueryParser::FLAG_DEFAULT | Xapian::QueryParser::FLAG_PARTIAL));
                 
 		i_free(s);
                 delete(qp);
@@ -305,23 +305,30 @@ class XQuerySet
 	}
 };
 
-class XHeaderTerm
+class XNGram
 {
-	public:
-		long size,partial,full,maxlength;
-		char ** data;
+	private:
+		long partial,full,hardlimit;
+		const char * prefix;
 		bool onlyone;
+
+	public:
+		char ** data;
+		long size,maxlength;
   
-        XHeaderTerm(long p, long f, bool o) 
+        XNGram(long p, long f, const char * pre) 
 	{ 
 		partial=p; full=f; 
 		size=0; 
 		maxlength=0;
 		data=NULL; 
-		onlyone=o;
+		prefix=pre;
+		hardlimit=XAPIAN_TERM_SIZELIMIT-strlen(prefix);
+		onlyone=false;
+		if(strcmp(prefix,"XMID")==0) onlyone=true;
 	}
 
-        ~XHeaderTerm() 
+        ~XNGram() 
 	{ 
 		if (size>0) 
 		{ 
@@ -373,12 +380,6 @@ class XHeaderTerm
 
 		if(l<partial) return;
 
-                if(onlyone)
-                {
-                        add_stem(d);
-                        return;
-                }
-
 		i = d->indexOf(" ");
 
 		if(i>0)
@@ -390,6 +391,12 @@ class XHeaderTerm
 			d->trim();
 			l=d->length();
 		}
+
+		if(onlyone)
+		{
+			add_stem(d);
+			return;
+		}
 	
 		for(i=0;i<=l-partial;i++)
 		{
@@ -413,7 +420,16 @@ class XHeaderTerm
 		std::string s;
 		d->toUTF8String(s);
 		l  = s.length();
-		if(l>XAPIAN_TERM_SIZELIMIT) return;
+		if(l>hardlimit)
+		{
+			icu::UnicodeString * r = new icu::UnicodeString(*d,1);
+			add_stem(r);
+			delete(r);
+			r = new icu::UnicodeString(*d,0,d->length()-1);
+			add_stem(r);
+			delete(r);
+			return;
+		}
 
 		char * s2 = i_strdup(s.c_str());
  
@@ -728,27 +744,27 @@ bool fts_backend_xapian_index_hdr(Xapian::WritableDatabase * dbx, uint uid, cons
 
 		if(i>=HDRS_NB) return true;
 		const char * h=hdrs_xapian[i];
-		{
-			XHeaderTerm * xhs = new XHeaderTerm(p,f,strcmp(h,"XMID")==0);
-	                xhs->add(data);
-	
-			char *t = (char*)i_malloc(sizeof(char)*(xhs->maxlength+6));
 		
-			for(i=0;i<xhs->size;i++)
+		XNGram * ngram = new XNGram(p,f,h);
+	        ngram->add(data);
+	
+		char *t = (char*)i_malloc(sizeof(char)*(ngram->maxlength+6));
+	
+		for(i=0;i<ngram->size;i++)
+		{
+			snprintf(t,ngram->maxlength+6,"%s%s",h,ngram->data[i]);
+			try
 			{
-				snprintf(t,xhs->maxlength+6,"%s%s",h,xhs->data[i]);
-				try
-				{
-					doc.add_term(t);
-				}
-				catch(Xapian::Error e)
-				{
-					i_error("FTS Xapian: %s",e.get_msg().c_str());
-				}
+				doc.add_term(t);
+			}
+			catch(Xapian::Error e)
+			{
+				i_error("FTS Xapian: %s",e.get_msg().c_str());
 			}
-			i_free(t);
-			delete(xhs);
 		}
+		i_free(t);
+		delete(ngram);
+		
 		dbx->replace_document(docid,doc);
 	    	return true;
 	}
@@ -760,7 +776,7 @@ bool fts_backend_xapian_index_hdr(Xapian::WritableDatabase * dbx, uint uid, cons
 	return false;
 }
 
-bool fts_backend_xapian_index_text(Xapian::WritableDatabase * dbx,uint uid, const char * field, const char * data)
+bool fts_backend_xapian_index_text(Xapian::WritableDatabase * dbx,uint uid, const char * field, const char * data,long p, long f)
 {
 	try
         {
@@ -788,9 +804,11 @@ bool fts_backend_xapian_index_text(Xapian::WritableDatabase * dbx,uint uid, cons
 		delete(result);
 		delete(xq);
 
+		Xapian::Document doc2;
 		Xapian::TermGenerator termgenerator;
-		termgenerator.set_stemmer(Xapian::Stem("en"));
-		termgenerator.set_document(doc);
+		Xapian::Stem stem("en");
+		termgenerator.set_stemmer(stem);
+		termgenerator.set_document(doc2);
 	
 		const char * h;
                 if(strcmp(field,"subject")==0) 
@@ -802,8 +820,39 @@ bool fts_backend_xapian_index_text(Xapian::WritableDatabase * dbx,uint uid, cons
 			h="XBDY";
 		}
 		std::string d(data);
+		termgenerator.set_stemming_strategy(Xapian::TermGenerator::STEM_ALL);
 		termgenerator.index_text(d, 1, h);
-		
+	
+		long n= doc2.termlist_count();
+		Xapian::TermIterator ti = doc2.termlist_begin();
+		XNGram * ngram = new XNGram(p,f,h);
+		while(n>0)
+		{
+			const std::string s = *ti;
+			if(s.compare(0,strlen(h),h)==0)
+			{
+				ngram->add(s.c_str()+strlen(h));
+			}
+			ti++;
+			n--;
+		}
+		if(verbose>1) i_info("NGRAM(%s,%s) %ld max=%ld",field,h,ngram->size,ngram->maxlength);
+		char *t = (char*)i_malloc(sizeof(char)*(ngram->maxlength+6));
+		for(n=0;n<ngram->size;n++)
+                {
+                        snprintf(t,ngram->maxlength+6,"%s%s",h,ngram->data[n]);
+                        try
+                        {
+                                doc.add_term(t);
+                        }
+                        catch(Xapian::Error e)
+                        {
+                                i_error("FTS Xapian: %s",e.get_msg().c_str());
+                        }
+                }
+                i_free(t);
+                delete(ngram);
+
                 dbx->replace_document(docid,doc);
           }
           catch(Xapian::Error e)
